/*!
 * @file stepper_motor_controller.ino
 *
 * This is the documentation for satnogs rotator controller firmware
 * for stepper motors configuration. The board (PCB) is placed in
 * <a href="https://gitlab.com/librespacefoundation/satnogs/satnogs-rotator-controller">
 * satnogs-rotator-controller </a> and is for releases:
 * v2.0
 * v2.1
 * v2.2
 * <a href="https://wiki.satnogs.org/SatNOGS_Rotator_Controller"> wiki page </a>
 *
 * @section dependencies Dependencies
 *
 * This firmware depends on <a href="http://www.airspayce.com/mikem/arduino/AccelStepper/index.htmlhttp://www.airspayce.com/mikem/arduino/AccelStepper/index.html">
 * AccelStepper library</a> being present on your system. Please make sure you
 * have installed the latest version before using this firmware.
 *
 * @section license License
 *
 * Licensed under the GPLv3.
 *
 */

/*
* AZ motor: Clockwise for raising Azimuth
* EL motor: Clockwise for raising Elevation
*/

/* Fichier gpredict_wifi pour connexion directe, pour carte Wemos D1 Mini Pro */

#define SAMPLE_TIME        0.1   ///< Control loop in s
#define BELT_RATIO         2.5   ///< Belt ratio of rotator - Poulies 30T et 12T
#define GEAR_REDUCER       23    ///< Gear ratio of rotator gear box
#define MIN_PULSE_WIDTH    20    ///< In microsecond for AccelStepper
#define MAX_SPEED          1200  ///< In steps/s, consider the microstep
#define MAX_ACCELERATION   600   ///< In steps/s^2, consider the microstep / moitié MAX_SPEED pas mal !
#define SPR_AZ             1210  ///< Azimuth Step Per Revolution, consider the microstep
#define SPR_EL             1135  ///< Elevation Step Per Revolution, consider the microstep
#define MIN_AZ_ANGLE       0     ///< Minimum angle of azimuth
#define MAX_AZ_ANGLE       360   ///< Maximum angle of azimuth
#define MIN_EL_ANGLE       0     ///< Minimum angle of elevation
#define MAX_EL_ANGLE       180   ///< Maximum angle of elevation
#define DEFAULT_HOME_STATE LOW   ///< Change to LOW according to Home sensor
#define HOME_DELAY         6000  ///< Time for homing Deceleration in millisecond

#include <AccelStepper.h>
#include "globals.h"
#include "easycomm_serial.h"
#include "gpredict_wifi.h"
#include "rotator_pins.h"
#include "endstop.h"

// Timeout reboot si perte Wifi
unsigned long delay_reboot = 300000; // 300 seconds delay
unsigned long previous_time = 0;

uint32_t t_run = 0; // run time of uC
easycomm commSerial;
gpredict commWifi;
AccelStepper stepper_az(1, AZ_STEP, AZ_DIR);
AccelStepper stepper_el(1, EL_STEP, EL_DIR);
endstop switch_az(SW_AZ, DEFAULT_HOME_STATE), switch_el(SW_EL, DEFAULT_HOME_STATE);

enum _rotator_error homing(int32_t seek_az, int32_t seek_el);
int32_t deg2step(float deg);
float step2deg(int32_t step);

void setup() {

    pinMode(WIFI_LED, OUTPUT);

    // Homing switch
    switch_az.init();
    switch_el.init();

    // Serial Communication
    commSerial.easycomm_init();

    // Wifi Communication
    commWifi.gpredict_init();

    // Stepper Motor setup
    stepper_az.setEnablePin(MOTOR_EN);
    stepper_az.setPinsInverted(false, false, true);
    stepper_az.enableOutputs();
    stepper_az.setMaxSpeed(MAX_SPEED);
    stepper_az.setAcceleration(MAX_ACCELERATION);
    stepper_az.setMinPulseWidth(MIN_PULSE_WIDTH);

    stepper_el.setPinsInverted(false, false, true);
    stepper_el.enableOutputs();
    stepper_el.setMaxSpeed(MAX_SPEED);
    stepper_el.setAcceleration(MAX_ACCELERATION);
    stepper_el.setMinPulseWidth(MIN_PULSE_WIDTH);
}

void loop() {

    // Get end stop status
    rotator.switch_az = switch_az.get_state();
    rotator.switch_el = switch_el.get_state();

    // Run easycomm implementation
    //commSerial.easycomm_proc();

    // Run GPredict implementation
    commWifi.gpredict_proc();

    // Get position of both axis
    control_az.input = step2deg(stepper_az.currentPosition(), SPR_AZ);
    control_el.input = step2deg(stepper_el.currentPosition(), SPR_EL);

    // Check rotator status
    if (rotator.rotator_status != error) {
        if (rotator.homing_flag == false) {
            // Check home flag
            rotator.control_mode = position;
            // Homing
            rotator.rotator_error = homing(deg2step(-MAX_AZ_ANGLE, SPR_AZ),
                                           deg2step(-MAX_EL_ANGLE, SPR_EL));
            if (rotator.rotator_error == no_error) {
                // No error
                rotator.rotator_status = idle;
                rotator.homing_flag = true;
            } else {
                // Error
                rotator.rotator_status = error;
                rotator.rotator_error = homing_error;
            }
        } else {
            // Control Loop
            stepper_az.moveTo(deg2step(control_az.setpoint, SPR_AZ));
            stepper_el.moveTo(deg2step(control_el.setpoint, SPR_EL));
            rotator.rotator_status = pointing;
            // Move azimuth and elevation motors
            stepper_az.run();
            stepper_el.run();
            // Idle rotator
            if (stepper_az.distanceToGo() == 0 && stepper_el.distanceToGo() == 0) {
                rotator.rotator_status = idle;
            }
        }
    } else {
        // Error handler, stop motors and disable the motor driver
        stepper_az.stop();
        stepper_az.disableOutputs();
        stepper_el.stop();
        stepper_el.disableOutputs();
        if (rotator.rotator_error != homing_error) {
            // Reset error according to error value
            rotator.rotator_error = no_error;
            rotator.rotator_status = idle;
        }
    }

        // Reboot si perte Wifi
    unsigned long current_time = millis();

    if (WiFi.status() == WL_CONNECTED) {
        previous_time = current_time;
    }
    else if (current_time - previous_time > delay_reboot) {
        ESP.restart();
    }
}

/**************************************************************************/
/*!
    @brief    Move both axis with one direction in order to find home position,
              end-stop switches
    @param    seek_az
              Steps to find home position for azimuth axis
    @param    seek_el
              Steps to find home position for elevation axis
    @return   _rotator_error
*/
/**************************************************************************/
enum _rotator_error homing(int32_t seek_az, int32_t seek_el) {
    bool isHome_az = false;
    bool isHome_el = false;

    // Move motors to "seek" position
    stepper_az.moveTo(seek_az);
    stepper_el.moveTo(seek_el);

    // Homing loop
    while (isHome_az == false || isHome_el == false) {
        // Update WDT
        ESP.wdtFeed();
        if (switch_az.get_state() == true && !isHome_az) {
            // Find azimuth home
            stepper_az.moveTo(stepper_az.currentPosition());
            isHome_az = true;
        }
        if (switch_el.get_state() == true && !isHome_el) {
            // Find elevation home
            stepper_el.moveTo(stepper_el.currentPosition());
            isHome_el = true;
        }
        // Check if the rotator goes out of limits or something goes wrong (in mechanical)
        if ((stepper_az.distanceToGo() == 0 && !isHome_az) ||
            (stepper_el.distanceToGo() == 0 && !isHome_el)){
            return homing_error;
        }
        // Move motors to "seek" position
        stepper_az.run();
        stepper_el.run();
    }
    // Delay to Deccelerate and homing, to complete the movements
    uint32_t time = millis();
    while (millis() - time < HOME_DELAY) {
        ESP.wdtFeed();
        stepper_az.run();
        stepper_el.run();
    }

    // Park bd Picpus : 45° / 90°
    // Était 0/0 initialement
    stepper_az.setCurrentPosition(45);
    stepper_el.setCurrentPosition(90);
    control_az.setpoint = 45;
    control_el.setpoint = 90;

    return no_error;
}

/**************************************************************************/
/*!
    @brief    Convert degrees to steps according to step/revolution, rotator
              gear box ratio and microstep
    @param    deg
              Degrees in float format
    @return   Steps for stepper motor driver, int32_t
*/
/**************************************************************************/
int32_t deg2step(float deg, int spr) {
    return (BELT_RATIO * GEAR_REDUCER * spr * deg / 360);
}

/**************************************************************************/
/*!
    @brief    Convert steps to degrees according to step/revolution, rotator
              gear box ratio and microstep
    @param    step
              Steps in int32_t format
    @return   Degrees in float format
*/
/**************************************************************************/
float step2deg(int32_t step, int spr) {
    return (360.00 * step / (spr * BELT_RATIO * GEAR_REDUCER));
}
